import os

from savvihub.api.file_object import UploadableFileObject, UploadableS3MultipartFileObject
from savvihub.common.utils import calculate_crc32c


def default_hooks():
    def fn(resp, **kwargs):
        resp.raise_for_status()
    return {
        'response': fn,
    }

class Uploader:
    @classmethod
    def get_files_to_upload(cls, local_base_path, hashmap=None):
        local_base_path = local_base_path.rstrip('/')
        results = []
        for root, dirs, files in os.walk(local_base_path):
            for name in files:
                name = os.path.join(root, name)
                name = name[len(local_base_path) + 1:] if name.startswith(local_base_path) else name
                if hashmap and hashmap[name] == calculate_crc32c(os.path.join(local_base_path, name)):
                    continue
                results.append(name)
        return results

    @classmethod
    def get_hashmap(cls, local_base_path):
        files = cls.get_files_to_upload(local_base_path)
        hashmap = dict()
        for file in files:
            path = os.path.join(local_base_path, file)
            hashmap[file] = calculate_crc32c(path)
        return hashmap

    @classmethod
    def upload(cls, client, local_path, volume_id, remote_path, progressable=None):
        file_object = client.volume_file_create(volume_id, remote_path, is_dir=False)

        mpu = UploadableS3MultipartFileObject(
            file_object.upload_url.federation_token.bucket,
            file_object.upload_url.federation_token.key,
            local_path,
            file_object.upload_url.federation_token.token,
        )
        mpu.abort_all()
        mpu_id = mpu.create()
        parts = mpu.upload(mpu_id)
        mpu.complete(mpu_id, parts)
        resp = client.volume_file_uploaded(volume_id, remote_path)
        return resp

    @classmethod
    def bulk_upload(cls, client, local_base_path, local_file_paths, volume_id, remote_base_path, *, progressable=None):
        # TODO: parallel upload
        if len(local_file_paths) <= 0:
            return

        file_objects = []
        for local_file_path in local_file_paths:
            file_objects.append(client.volume_file_create(
                volume_id,
                os.path.join(remote_base_path, local_file_path),
                is_dir=False,
                # hooks=default_hooks()
            ))

        for i, file_object in enumerate(file_objects):
            mpu = UploadableS3MultipartFileObject(
                file_object.upload_url.federation_token.bucket,
                file_object.upload_url.federation_token.key,
                os.path.join(local_base_path, local_file_paths[i]),
                file_object.upload_url.federation_token.token,
            )
            mpu.abort_all()
            mpu_id = mpu.create()
            parts = mpu.upload(mpu_id)
            mpu.complete(mpu_id, parts)

        for local_file_path in local_file_paths:
            client.volume_file_uploaded(
                volume_id,
                os.path.join(remote_base_path, local_file_path),
                # hooks=default_hooks(),
            )

        return file_objects



from urllib.parse import urlparse

import typer
from terminaltables import AsciiTable

from savvihub.cli.commands.dataset_files import dataset_files_app
from savvihub.cli.constants import (
    DATASET_PATH_PARSE_SCHEME_S3_WITH_SEPERATOR,
    DATASET_PATH_PARSE_SCHEME_GS_WITH_SEPERATOR,
    DATASET_PATH_PARSE_SCHEME_S3,
    DATASET_PATH_PARSE_SCHEME_GS,
    DATASET_SOURCE_TYPE_AWS_S3,
    DATASET_SOURCE_TYPE_GCP_GS,
    WEB_HOST,
)
from savvihub.cli.exceptions import ExitException
from savvihub.cli.inputs.organization import organization_name_option
from savvihub.cli.typer import Typer, Context
from savvihub.cli.utils import parse_dataset

dataset_app = Typer()
dataset_app.add_typer(dataset_files_app, name='files')


@dataset_app.callback()
def main():
    """
    Manage the collection of data
    """
    return


@dataset_app.command(user_required=True)
def list(
    ctx: Context,
    organization_name: str = organization_name_option,
):
    """
    Show datasets
    """
    client = ctx.authenticated_client
    dataset_list = client.dataset_list(organization_name).results
    if len(dataset_list) == 0:
        typer.echo(f'No datasets found in `{organization_name}` organization.')
    else:
        table = AsciiTable([
            ['ORGANIZATION/NAME', 'SOURCE TYPE'],
            *[[f'{d.organization.name}/{d.name}', d.source.type] for d in dataset_list],
        ])
        table.inner_column_border = False
        table.inner_heading_row_border = False
        table.inner_footing_row_border = False
        table.outer_border = False

        typer.echo(table.table)


@dataset_app.command(user_required=True)
def create(
    ctx: Context,
    organization_name: str = organization_name_option,
    dataset_name: str = typer.Argument(...),
    is_version_enabled: bool = typer.Option(False, "-v", "--versioning", help="Enable or disable versioning."),
    path_arg: str = typer.Option(None, "-u", "--url", help="S3 or GoogleStorage url starting with s3:// or gs://."),
    aws_role_arn: str = typer.Option(None, "--aws-role-arn", help="Required to create S3 dataset."),
    version_path_arg: str = typer.Option(None, "-vu", "--version-url", help="Required to version S3 or GoogleStorage dataset."),
    description: str = typer.Option(None, "-m", help="Dataset description."),
):
    """
    Create a dataset
    """
    client = ctx.authenticated_client

    if path_arg:
        if not (path_arg.startswith(DATASET_PATH_PARSE_SCHEME_S3_WITH_SEPERATOR)
                or path_arg.startswith(DATASET_PATH_PARSE_SCHEME_GS_WITH_SEPERATOR)):
            raise ExitException(f"path should start with \"s3://\" or \"gs://\"")

        r = urlparse(path_arg)
        if is_version_enabled:
            if not version_path_arg:
                raise ExitException("Versioning bucket path is required to version S3 or GoogleStorage.")
            if r.scheme == DATASET_PATH_PARSE_SCHEME_S3:
                if not version_path_arg.startswith(DATASET_PATH_PARSE_SCHEME_S3_WITH_SEPERATOR):
                    raise ExitException(f"path should start with \"s3://\"")
                if path_arg.startswith(version_path_arg):
                    raise ExitException("data bucket path shouldn't be a child directory of version bucket path")
                if version_path_arg.startswith(path_arg):
                    raise ExitException("version bucket path shouldn't be a child directory of data bucket path")
            if r.scheme == DATASET_PATH_PARSE_SCHEME_GS:
                raise ExitException("Currently versioning is not allowed for GoogleStorage.")
                """
                if not version_path_arg.startswith(DATASET_PATH_PARSE_SCHEME_GS_WITH_SEPERATOR):
                    raise ExitException(f"path should start with \"gs://\"")
                """

        if r.scheme == DATASET_PATH_PARSE_SCHEME_GS:
            dataset = client.dataset_gs_create(organization_name, dataset_name, False, description, path_arg)
        elif r.scheme == DATASET_PATH_PARSE_SCHEME_S3:
            if not aws_role_arn:
                raise ExitException("AWS Role ARN is required for S3 users")
            dataset = client.dataset_s3_create(
                organization_name,
                dataset_name,
                False,
                description,
                path_arg,
                aws_role_arn,
                is_version_enabled,
                version_path_arg,
            )
        else:
            raise ExitException("Only Google Cloud Storage and Amazon S3 are supported at the moment.")
    else:
        dataset = client.dataset_create(organization_name, dataset_name, is_version_enabled, description)

    if not dataset:
        return

    if dataset_name != dataset.name:
        typer.echo(f'Duplicate dataset name: {dataset_name}')

    typer.echo(
        f'Dataset {dataset.name} is created.\n'
        f'Full dataset info at:\n'
        f'    {WEB_HOST}/{dataset.organization.name}/datasets/{dataset.name}\n'
    )


@dataset_app.command(user_required=True)
def describe(
    ctx: Context,
    organization_name: str = organization_name_option,
    dataset_name: str = typer.Argument(...),
):
    """
    Describe the dataset information in detail
    """
    organization_name_override, dataset_name, _ = parse_dataset(dataset_name)
    if organization_name_override:
        organization_name = organization_name_override

    dataset = ctx.authenticated_client.dataset_read(organization_name, dataset_name)
    typer.echo(
        f'Name: {dataset.name}\n'
        f'Versioning: {dataset.is_version_enabled}\n'
        f'Volume ID: {dataset.volume_id}\n'
        f'Organization: {dataset.organization.name}'
    )

    source = dataset.source
    if source.type == DATASET_SOURCE_TYPE_AWS_S3:
        typer.echo(f'Source: s3://{source.bucket_name}/{source.path}')
    elif source.type == DATASET_SOURCE_TYPE_GCP_GS:
        typer.echo(f'Source: gs://{source.bucket_name}/{source.path}')

    typer.echo(
        '\n'
        f'Full dataset info at:\n'
        f'    {WEB_HOST}/{dataset.organization.name}/datasets/{dataset.name}\n'
    )



import os
import urllib.parse
from pathlib import Path

import inquirer
import paramiko
import typer

from savvihub.cli.commands.ssh_keys import ssh_keys_app
from savvihub.cli.exceptions import ExitException
from savvihub.cli.inputs.organization import organization_name_option
from savvihub.cli.inputs.ssh import ssh_private_key_path_callback
from savvihub.cli.typer import Typer, Context
from savvihub.common.utils import parse_time_to_ago

ssh_app = Typer()
ssh_app.add_typer(ssh_keys_app, name='keys')


@ssh_app.callback()
def main():
    """
    Manage ssh keys and connect to workspaces
    """


@ssh_app.command(user_required=True)
def connect(
    ctx: Context,
    organization_name: str = organization_name_option,
    ssh_private_key_path: str = typer.Option(None, '--key-path', callback=ssh_private_key_path_callback, help='SSH private key path.'),
):
    """
    Connect to a running workspace.
    """
    workspace_list_response = ctx.authenticated_client.workspace_list(organization_name, 'running')
    running_workspaces = workspace_list_response.results
    if len(running_workspaces) == 0:
        raise ExitException('There is no running workspace.')

    if len(running_workspaces) == 1:
        workspace = running_workspaces[0]
    else:
        workspace = inquirer.prompt([inquirer.List(
            'question',
            message='Select workspace',
            choices=[(f'{w.name} (created {parse_time_to_ago(w.created_dt)})', w) for w in running_workspaces],
        )], raise_keyboard_interrupt=True).get('question')

    ssh_endpoint = urllib.parse.urlparse(workspace.endpoints.ssh.endpoint)
    ssh_private_key_option = f' -i {ssh_private_key_path}' if ssh_private_key_path else ''
    os.system(f'ssh -p {ssh_endpoint.port}{ssh_private_key_option} {ctx.user.username}@{ssh_endpoint.hostname}')


@ssh_app.command(user_required=True)
def vscode(
    ctx: Context,
    organization_name: str = organization_name_option,
    ssh_private_key_path: str = typer.Option(None, '--key-path', callback=ssh_private_key_path_callback,
                                             help='SSH private key path.'),
):
    """
    Update .ssh/config file for VSCode Remote-SSH plugin
    """
    workspace_list_response = ctx.authenticated_client.workspace_list(organization_name, 'running')
    running_workspaces = workspace_list_response.results
    if len(running_workspaces) == 0:
        raise ExitException('There is no running workspace.')

    ssh_config = paramiko.SSHConfig()
    ssh_config.parse(open(f'{Path.home()}/.ssh/config'))
    hostname_set = ssh_config.get_hostnames()

    for workspace in running_workspaces:
        hostname = f'{workspace.name}-{int(workspace.created_dt.timestamp())}'
        if hostname in hostname_set:
            continue

        ssh_endpoint = urllib.parse.urlparse(workspace.endpoints.ssh.endpoint)
        config_value = f'''
Host {hostname}
    User {ctx.user.username}
    Hostname {ssh_endpoint.hostname}
    Port {ssh_endpoint.port}
    StrictHostKeyChecking accept-new
    CheckHostIP no
'''
        if ssh_private_key_path:
            config_value += f'    IdentityFile {ssh_private_key_path}\n'

        with open(f'{Path.home()}/.ssh/config', 'a') as f:
            f.write(config_value)

    typer.echo(f'Successfully updated {Path.home()}/.ssh/config')



import os
from pathlib import Path
from typing import Optional

import inquirer
from sshpubkeys import SSHKey, InvalidKeyError

from savvihub.cli.exceptions import ExitException
from savvihub.cli.typer import Context


def ssh_public_key_value_callback(ctx: Context, key_path: str) -> str:
    if not key_path:
        key_path = inquirer.prompt([inquirer.Text(
            'question',
            message='SSH public key path',
            default=f'{Path.home()}/.ssh/id_ed25519.pub',
        )], raise_keyboard_interrupt=True).get('question')

    try:
        with open(key_path, 'r') as f:
            ssh_public_key_value = f.read()
    except FileNotFoundError:
        raise ExitException('Key file not found.')

    ssh = SSHKey(ssh_public_key_value, strict=True)
    try:
        ssh.parse()
    except InvalidKeyError as e:
        raise ExitException('Invalid key:', e)
    except NotImplementedError as e:
        raise ExitException('Invalid key type:', e)

    ctx.store['ssh_key_name'] = ssh.comment
    ctx.store['ssh_key_filename'] = os.path.basename(key_path)
    return ssh_public_key_value


def ssh_key_name_callback(ctx: Context, name: str) -> str:
    if name:
        return name

    return inquirer.prompt([inquirer.Text(
        'question',
        message='SSH public key name',
        default=ctx.store.get('ssh_key_name'),
    )], raise_keyboard_interrupt=True).get('question')


def ssh_private_key_path_callback(ctx: Context, key_path: str) -> Optional[str]:
    if key_path:
        return key_path

    ssh_keys_response = ctx.authenticated_client.ssh_key_list()
    if len(ssh_keys_response.ssh_keys) == 0:
        raise ExitException('At least one ssh public key should be added.\n'
                            'Please run `savvihub ssh keys add`.')

    home = Path.home()

    for ssh_key in ssh_keys_response.ssh_keys:
        key_path = f'{home}/.ssh/{ssh_key.filename.rstrip(".pub")}'
        if os.path.exists(key_path):
            return key_path

    for key_path in (f'{home}/.ssh/id_rsa', f'{home}/.ssh/id_ed25519'):
        if os.path.exists(key_path):
            return key_path

    return None



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class ModelOrganizationEdges(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'billing_histories': 'list[ModelBillingHistory]',
        'default_kernel_cluster': 'ModelKernelCluster',
        'default_storage': 'ModelStorage',
        'default_volume': 'ModelVolume',
        'notification_configs': 'list[ModelNotificationConfig]',
        'primary_owner': 'ModelUser',
        'user_organization': 'list[ModelUserOrganization]'
    }

    attribute_map = {
        'billing_histories': 'billing_histories',
        'default_kernel_cluster': 'default_kernel_cluster',
        'default_storage': 'default_storage',
        'default_volume': 'default_volume',
        'notification_configs': 'notification_configs',
        'primary_owner': 'primary_owner',
        'user_organization': 'user_organization'
    }

    def __init__(self, billing_histories=None, default_kernel_cluster=None, default_storage=None, default_volume=None, notification_configs=None, primary_owner=None, user_organization=None, local_vars_configuration=None):  # noqa: E501
        """ModelOrganizationEdges - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._billing_histories = None
        self._default_kernel_cluster = None
        self._default_storage = None
        self._default_volume = None
        self._notification_configs = None
        self._primary_owner = None
        self._user_organization = None
        self.discriminator = None

        if billing_histories is not None:
            self.billing_histories = billing_histories
        if default_kernel_cluster is not None:
            self.default_kernel_cluster = default_kernel_cluster
        if default_storage is not None:
            self.default_storage = default_storage
        if default_volume is not None:
            self.default_volume = default_volume
        if notification_configs is not None:
            self.notification_configs = notification_configs
        if primary_owner is not None:
            self.primary_owner = primary_owner
        if user_organization is not None:
            self.user_organization = user_organization

    @property
    def billing_histories(self):
        """Gets the billing_histories of this ModelOrganizationEdges.  # noqa: E501


        :return: The billing_histories of this ModelOrganizationEdges.  # noqa: E501
        :rtype: list[ModelBillingHistory]
        """
        return self._billing_histories

    @billing_histories.setter
    def billing_histories(self, billing_histories):
        """Sets the billing_histories of this ModelOrganizationEdges.


        :param billing_histories: The billing_histories of this ModelOrganizationEdges.  # noqa: E501
        :type billing_histories: list[ModelBillingHistory]
        """

        self._billing_histories = billing_histories

    @property
    def default_kernel_cluster(self):
        """Gets the default_kernel_cluster of this ModelOrganizationEdges.  # noqa: E501


        :return: The default_kernel_cluster of this ModelOrganizationEdges.  # noqa: E501
        :rtype: ModelKernelCluster
        """
        return self._default_kernel_cluster

    @default_kernel_cluster.setter
    def default_kernel_cluster(self, default_kernel_cluster):
        """Sets the default_kernel_cluster of this ModelOrganizationEdges.


        :param default_kernel_cluster: The default_kernel_cluster of this ModelOrganizationEdges.  # noqa: E501
        :type default_kernel_cluster: ModelKernelCluster
        """

        self._default_kernel_cluster = default_kernel_cluster

    @property
    def default_storage(self):
        """Gets the default_storage of this ModelOrganizationEdges.  # noqa: E501


        :return: The default_storage of this ModelOrganizationEdges.  # noqa: E501
        :rtype: ModelStorage
        """
        return self._default_storage

    @default_storage.setter
    def default_storage(self, default_storage):
        """Sets the default_storage of this ModelOrganizationEdges.


        :param default_storage: The default_storage of this ModelOrganizationEdges.  # noqa: E501
        :type default_storage: ModelStorage
        """

        self._default_storage = default_storage

    @property
    def default_volume(self):
        """Gets the default_volume of this ModelOrganizationEdges.  # noqa: E501


        :return: The default_volume of this ModelOrganizationEdges.  # noqa: E501
        :rtype: ModelVolume
        """
        return self._default_volume

    @default_volume.setter
    def default_volume(self, default_volume):
        """Sets the default_volume of this ModelOrganizationEdges.


        :param default_volume: The default_volume of this ModelOrganizationEdges.  # noqa: E501
        :type default_volume: ModelVolume
        """

        self._default_volume = default_volume

    @property
    def notification_configs(self):
        """Gets the notification_configs of this ModelOrganizationEdges.  # noqa: E501


        :return: The notification_configs of this ModelOrganizationEdges.  # noqa: E501
        :rtype: list[ModelNotificationConfig]
        """
        return self._notification_configs

    @notification_configs.setter
    def notification_configs(self, notification_configs):
        """Sets the notification_configs of this ModelOrganizationEdges.


        :param notification_configs: The notification_configs of this ModelOrganizationEdges.  # noqa: E501
        :type notification_configs: list[ModelNotificationConfig]
        """

        self._notification_configs = notification_configs

    @property
    def primary_owner(self):
        """Gets the primary_owner of this ModelOrganizationEdges.  # noqa: E501


        :return: The primary_owner of this ModelOrganizationEdges.  # noqa: E501
        :rtype: ModelUser
        """
        return self._primary_owner

    @primary_owner.setter
    def primary_owner(self, primary_owner):
        """Sets the primary_owner of this ModelOrganizationEdges.


        :param primary_owner: The primary_owner of this ModelOrganizationEdges.  # noqa: E501
        :type primary_owner: ModelUser
        """

        self._primary_owner = primary_owner

    @property
    def user_organization(self):
        """Gets the user_organization of this ModelOrganizationEdges.  # noqa: E501


        :return: The user_organization of this ModelOrganizationEdges.  # noqa: E501
        :rtype: list[ModelUserOrganization]
        """
        return self._user_organization

    @user_organization.setter
    def user_organization(self, user_organization):
        """Sets the user_organization of this ModelOrganizationEdges.


        :param user_organization: The user_organization of this ModelOrganizationEdges.  # noqa: E501
        :type user_organization: list[ModelUserOrganization]
        """

        self._user_organization = user_organization

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ModelOrganizationEdges):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ModelOrganizationEdges):
            return True

        return self.to_dict() != other.to_dict()



import base64
import json
import os

import boto3
import requests
from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper

from savvihub.common.utils import read_in_chunks


class UploadableFileObject:
    def __init__(self, url, base_path, path):
        self.url = url
        self.base_path = base_path
        self.full_path = os.path.join(base_path, path)
        self.path = path

    def upload_chunks(self, *, callback=None):
        return read_in_chunks(self.full_path, callback=callback)

    def upload_hooks(self, *, callback=None):
        def fn(resp, **kwargs):
            if resp.status_code != 200:
                print(f'Upload for {resp.request.url} failed. Detail: {resp.data}')
        return {
            'response': fn,
        }

    def upload(self, session=requests.Session()):
        file_size = os.path.getsize(self.full_path)

        # send empty data when file is empty
        if os.stat(self.full_path).st_size == 0:
            future = session.put(
                self.url,
                data='',
                headers={'content-type': 'application/octet-stream'},
                hooks=self.upload_hooks(),
            )
            return future

        with open(self.full_path, "rb") as f:
            with tqdm(total=file_size, desc=self.path, unit="B", unit_scale=True, unit_divisor=1024) as t:
                wrapped_file = CallbackIOWrapper(t.update, f, "read")
                requests.put(self.url, data=wrapped_file)
        return

class UploadableS3MultipartFileObject:
    def __init__(self, bucket, key, local_path, token, part_size=int(15e6), verbose=False):
        self.bucket = bucket
        self.key = key
        self.path = local_path
        self.total_bytes = os.stat(local_path).st_size
        self.part_bytes = part_size
        self.token = token
        self.s3_client = self._get_s3_client_from_token()
        if verbose:
            boto3.set_stream_logger(name="botocore")

    def abort_all(self):
        mpus = self.s3_client.list_multipart_uploads(Bucket=self.bucket, Prefix=self.key)
        aborted = []
        print(f"Aborting {len(mpus)} uploads")
        if "Uploads" in mpus:
            for u in mpus["Uploads"]:
                upload_id = u["UploadId"]
                aborted.append(
                    self.s3_client.abort_multipart_upload(Bucket=self.bucket, Key=self.key, UploadId=upload_id))
        return aborted

    def create(self):
        mpu = self.s3_client.create_multipart_upload(Bucket=self.bucket, Key=self.key)
        mpu_id = mpu["UploadId"]
        return mpu_id

    def upload(self, mpu_id):
        parts = []
        uploaded_bytes = 0
        with open(self.path, "rb") as f:
            i = 1
            while True:
                data = f.read(self.part_bytes)
                if not len(data):
                    break
                part = self.s3_client.upload_part(
                    Body=data, Bucket=self.bucket, Key=self.key, UploadId=mpu_id, PartNumber=i)
                parts.append({"PartNumber": i, "ETag": part["ETag"]})
                uploaded_bytes += len(data)
                print(f"{self.path}: {uploaded_bytes} of {self.total_bytes} uploaded "
                      f"({(float(uploaded_bytes) / float(self.total_bytes) * 100.0):.3f}%)")
                i += 1
        return parts

    def complete(self, mpu_id, parts):
        result = self.s3_client.complete_multipart_upload(
            Bucket=self.bucket,
            Key=self.key,
            UploadId=mpu_id,
            MultipartUpload={"Parts": parts})
        return result

    def _get_s3_client_from_token(self):
        credentials = base64.b64decode(self.token).decode()
        credentials_dict = json.loads(credentials)
        return boto3.client(
            's3',
            aws_access_key_id=credentials_dict['AccessKeyId'],
            aws_secret_access_key=credentials_dict['SecretAccessKey'],
            aws_session_token=credentials_dict['SessionToken'],
        )


class DownloadableFileObject:
    def __init__(self, url, base_path, path, size=None):
        self.url = url
        self.full_path = os.path.join(base_path, path)
        self.size = size

    def download_hooks(self, *, callback=None):
        def fn(resp, **kwargs):
            os.makedirs(os.path.dirname(self.full_path), exist_ok=True)
            with open(self.full_path, 'wb') as f:
                for chunk in resp.iter_content(chunk_size=8192):
                    f.write(chunk)
                    if callback:
                        callback(chunk)
        return {
            'response': fn,
        }

    def download(self, session=requests.Session(), progressable=None):
        progress_callback = None
        if progressable and self.size:
            progress = progressable(length=self.size, label=self.full_path)
            progress_callback = lambda data: progress.update(len(data))

        future = session.get(
            self.url,
            stream=True,
            hooks=self.download_hooks(callback=progress_callback)
        )
        return future
